from datetime import datetime
from unittest import TestCase
from unittest.mock import MagicMock, Mock, PropertyMock

from requests import HTTPError, Response

from pycarlo.common.errors import InvalidSessionError
from pycarlo.core import Client
from pycarlo.features.ingestion.exceptions import IngestionError
from pycarlo.features.ingestion.models import (
    AssetField,
    AssetFreshness,
    AssetMetadata,
    AssetVolume,
    ColumnLineageField,
    ColumnLineageSourceField,
    LineageAssetRef,
    LineageEvent,
    QueryLogEntry,
    RelationalAsset,
    Tag,
)
from pycarlo.features.ingestion.service import IngestionService


def _dt(iso: str) -> datetime:
    """Parse ISO8601 string to datetime for tests."""
    return datetime.fromisoformat(iso.replace("Z", "+00:00"))


def _mock_client(scope: str = "Ingestion") -> MagicMock:
    client = MagicMock(spec=Client)
    type(client).session_scope = PropertyMock(return_value=scope)
    return client


class TestIngestionServiceInit(TestCase):
    def test_raises_without_scope(self):
        client = MagicMock(spec=Client)
        type(client).session_scope = PropertyMock(return_value=None)
        with self.assertRaises(InvalidSessionError):
            IngestionService(mc_client=client)

    def test_creates_with_scoped_client(self):
        client = _mock_client()
        svc = IngestionService(mc_client=client)
        assert svc._client is client


class TestExtractInvocationId(TestCase):
    def test_returns_invocation_id_when_present(self):
        assert (
            IngestionService.extract_invocation_id({"invocation_id": "test-invocation-id"})
            == "test-invocation-id"
        )

    def test_returns_none_when_response_missing(self):
        assert IngestionService.extract_invocation_id(None) is None

    def test_returns_none_when_value_is_not_string(self):
        assert IngestionService.extract_invocation_id({"invocation_id": 123}) is None


class TestSendMetadata(TestCase):
    def test_sends_correct_payload(self):
        client = _mock_client()
        client.make_request.return_value = {"status": "ok"}
        svc = IngestionService(mc_client=client)

        asset = RelationalAsset(
            type="TABLE",
            metadata=AssetMetadata(name="orders", database="analytics", schema="public"),
            fields=[AssetField(name="id", type="INTEGER")],
            volume=AssetVolume(row_count=1000),
            freshness=AssetFreshness(last_update_time="2026-03-02T10:00:00Z"),
        )

        result = svc.send_metadata(
            resource_uuid="res-001",
            resource_type="snowflake",
            events=[asset],
        )

        assert result == {"status": "ok"}
        client.make_request.assert_called_once()
        call_kwargs = client.make_request.call_args
        assert call_kwargs.kwargs["path"] == "/ingest/v1/metadata"
        assert call_kwargs.kwargs["method"] == "POST"

        body = call_kwargs.kwargs["body"]
        assert body["event_type"] == "METADATA"
        assert body["resource"]["uuid"] == "res-001"
        assert body["resource"]["resource_type"] == "snowflake"
        assert len(body["events"]) == 1

        relational_asset = body["events"][0]["relational_asset"]
        assert relational_asset["type"] == "TABLE"
        assert relational_asset["metadata"]["name"] == "orders"
        assert relational_asset["fields"] == [{"name": "id", "type": "INTEGER"}]
        assert relational_asset["volume"] == {"row_count": 1000}
        assert relational_asset["freshness"] == {
            "last_update_time": "2026-03-02T10:00:00Z",
        }

    def test_exposes_invocation_id_from_response(self):
        client = _mock_client()
        client.make_request.return_value = {"invocation_id": "metadata-invocation-id"}
        svc = IngestionService(mc_client=client)

        result = svc.send_metadata(
            resource_uuid="res-001",
            resource_type="snowflake",
            events=[
                RelationalAsset(
                    type="TABLE",
                    metadata=AssetMetadata(name="orders", database="analytics", schema="public"),
                )
            ],
        )

        assert IngestionService.extract_invocation_id(result) == "metadata-invocation-id"

    def test_sends_multiple_events(self):
        client = _mock_client()
        client.make_request.return_value = None
        svc = IngestionService(mc_client=client)

        events = [
            RelationalAsset(
                type="TABLE",
                metadata=AssetMetadata(name="t1", database="db", schema="s"),
            ),
            RelationalAsset(
                type="VIEW",
                metadata=AssetMetadata(name="v1", database="db", schema="s"),
                tags=[Tag(key="env", value="prod")],
            ),
        ]

        svc.send_metadata(
            resource_uuid="res-002",
            resource_type="bigquery",
            events=events,
        )

        body = client.make_request.call_args.kwargs["body"]
        assert len(body["events"]) == 2
        assert body["events"][0]["relational_asset"]["type"] == "TABLE"
        assert body["events"][1]["relational_asset"]["tags"] == [
            {"key": "env", "value": "prod"},
        ]

    def test_raises_on_empty_events(self):
        client = _mock_client()
        svc = IngestionService(mc_client=client)

        with self.assertRaises(ValueError, msg="At least one"):
            svc.send_metadata(
                resource_uuid="res-003",
                resource_type="snowflake",
                events=[],
            )
        client.make_request.assert_not_called()

    def test_wraps_http_error_as_ingestion_error(self):
        client = _mock_client()
        response = Mock(spec=Response)
        response.text = '{"error": "unauthorized"}'
        http_error = HTTPError(response=response)
        client.make_request.side_effect = http_error

        svc = IngestionService(mc_client=client)

        with self.assertRaises(IngestionError) as ctx:
            svc.send_metadata(
                resource_uuid="res-004",
                resource_type="snowflake",
                events=[
                    RelationalAsset(
                        type="TABLE",
                        metadata=AssetMetadata(name="t", database="d", schema="s"),
                    ),
                ],
            )
        assert "unauthorized" in str(ctx.exception)


class TestSendMetadataRaw(TestCase):
    def test_sends_raw_payload(self):
        client = _mock_client()
        client.make_request.return_value = {"status": "ok"}
        svc = IngestionService(mc_client=client)

        raw = {
            "event_type": "METADATA",
            "resource": {"uuid": "r1", "resource_type": "snowflake"},
            "events": [
                {
                    "relational_asset": {
                        "type": "TABLE",
                        "metadata": {"name": "t", "database": "d", "schema": "s"},
                    },
                },
            ],
        }

        result = svc.send_metadata_raw(payload=raw)

        assert result == {"status": "ok"}
        client.make_request.assert_called_once_with(
            path="/ingest/v1/metadata",
            method="POST",
            body=raw,
        )

    def test_wraps_http_error(self):
        client = _mock_client()
        response = Mock(spec=Response)
        response.text = "Bad Request"
        client.make_request.side_effect = HTTPError(response=response)

        svc = IngestionService(mc_client=client)

        with self.assertRaises(IngestionError):
            svc.send_metadata_raw(payload={"event_type": "METADATA"})


# ---------------------------------------------------------------
# send_lineage
# ---------------------------------------------------------------


def _table_event() -> LineageEvent:
    return LineageEvent(
        destination=LineageAssetRef(
            type="TABLE", database="analytics", schema="pub", name="summary"
        ),
        sources=[
            LineageAssetRef(type="TABLE", database="raw", schema="pub", name="orders"),
            LineageAssetRef(type="TABLE", database="raw", schema="pub", name="customers"),
        ],
    )


def _column_event() -> LineageEvent:
    return LineageEvent(
        destination=LineageAssetRef(type="TABLE", database="db", schema="s", name="dst"),
        sources=[
            LineageAssetRef(type="TABLE", database="db", schema="s", name="src", asset_id="src1"),
        ],
        fields=[
            ColumnLineageField(
                name="total",
                source_fields=[
                    ColumnLineageSourceField(asset_id="src1", field_name="amount"),
                ],
            ),
        ],
    )


class TestSendLineage(TestCase):
    def test_sends_correct_table_lineage_payload(self):
        client = _mock_client()
        client.make_request.return_value = {"status": "ok"}
        svc = IngestionService(mc_client=client)

        result = svc.send_lineage(
            resource_uuid="res-001",
            resource_type="snowflake",
            events=[_table_event()],
        )

        assert result == {"status": "ok"}
        client.make_request.assert_called_once()
        call_kwargs = client.make_request.call_args
        assert call_kwargs.kwargs["path"] == "/ingest/v1/lineage"
        assert call_kwargs.kwargs["method"] == "POST"

        body = call_kwargs.kwargs["body"]
        assert body["event_type"] == "LINEAGE"
        assert body["resource"]["uuid"] == "res-001"
        assert body["resource"]["resource_type"] == "snowflake"
        assert len(body["events"]) == 1
        assert body["events"][0]["destination"]["name"] == "summary"
        assert len(body["events"][0]["sources"]) == 2

    def test_exposes_invocation_id_from_response(self):
        client = _mock_client()
        client.make_request.return_value = {"invocation_id": "lineage-invocation-id"}
        svc = IngestionService(mc_client=client)

        result = svc.send_lineage(
            resource_uuid="res-001",
            resource_type="snowflake",
            events=[_table_event()],
        )

        assert IngestionService.extract_invocation_id(result) == "lineage-invocation-id"

    def test_sends_column_lineage_payload(self):
        client = _mock_client()
        client.make_request.return_value = {"status": "ok"}
        svc = IngestionService(mc_client=client)

        result = svc.send_lineage(
            resource_uuid="res-001",
            resource_type="snowflake",
            events=[_column_event()],
        )

        assert result == {"status": "ok"}
        body = client.make_request.call_args.kwargs["body"]
        assert body["event_type"] == "COLUMN_LINEAGE"
        assert "fields" in body["events"][0]
        assert body["events"][0]["fields"][0]["name"] == "total"

    def test_explicit_event_type_overrides(self):
        client = _mock_client()
        client.make_request.return_value = None
        svc = IngestionService(mc_client=client)

        svc.send_lineage(
            resource_uuid="res-001",
            resource_type="snowflake",
            events=[_column_event()],
            event_type="LINEAGE",
        )

        body = client.make_request.call_args.kwargs["body"]
        assert body["event_type"] == "LINEAGE"

    def test_sends_multiple_events(self):
        client = _mock_client()
        client.make_request.return_value = None
        svc = IngestionService(mc_client=client)

        svc.send_lineage(
            resource_uuid="res-002",
            resource_type="bigquery",
            events=[_table_event(), _table_event()],
        )

        body = client.make_request.call_args.kwargs["body"]
        assert len(body["events"]) == 2

    def test_raises_on_empty_events(self):
        client = _mock_client()
        svc = IngestionService(mc_client=client)

        with self.assertRaises(ValueError, msg="At least one"):
            svc.send_lineage(
                resource_uuid="res-003",
                resource_type="snowflake",
                events=[],
            )
        client.make_request.assert_not_called()

    def test_wraps_http_error_as_ingestion_error(self):
        client = _mock_client()
        response = Mock(spec=Response)
        response.text = '{"error": "unauthorized"}'
        client.make_request.side_effect = HTTPError(response=response)

        svc = IngestionService(mc_client=client)

        with self.assertRaises(IngestionError) as ctx:
            svc.send_lineage(
                resource_uuid="res-004",
                resource_type="snowflake",
                events=[_table_event()],
            )
        assert "unauthorized" in str(ctx.exception)


# ---------------------------------------------------------------
# send_query_logs
# ---------------------------------------------------------------


class TestSendQueryLogs(TestCase):
    def test_sends_correct_payload(self):
        client = _mock_client()
        client.make_request.return_value = {"status": "ok"}
        svc = IngestionService(mc_client=client)

        entry = QueryLogEntry(
            start_time=_dt("2026-03-02T10:00:00Z"),
            end_time=_dt("2026-03-02T10:00:05Z"),
            query_text="SELECT * FROM orders",
            query_id="q-123",
            returned_rows=100,
        )

        result = svc.send_query_logs(
            resource_uuid="res-001",
            resource_type="snowflake",
            events=[entry],
        )

        assert result == {"status": "ok"}
        client.make_request.assert_called_once()
        call_kwargs = client.make_request.call_args
        assert call_kwargs.kwargs["path"] == "/ingest/v1/querylogs"
        assert call_kwargs.kwargs["method"] == "POST"

        body = call_kwargs.kwargs["body"]
        assert body["event_type"] == "QUERY_LOG"
        assert body["resource"]["uuid"] == "res-001"
        assert body["resource"]["resource_type"] == "snowflake"
        assert len(body["events"]) == 1
        assert body["events"][0]["start_time"] == "2026-03-02T10:00:00Z"
        assert body["events"][0]["query_text"] == "SELECT * FROM orders"
        assert body["events"][0]["query_id"] == "q-123"
        assert body["events"][0]["returned_rows"] == 100

    def test_sends_multiple_events(self):
        client = _mock_client()
        client.make_request.return_value = None
        svc = IngestionService(mc_client=client)

        events = [
            QueryLogEntry(
                start_time=_dt("2026-03-02T10:00:00Z"),
                end_time=_dt("2026-03-02T10:00:01Z"),
                query_text=f"SELECT {i}",
            )
            for i in range(3)
        ]

        svc.send_query_logs(
            resource_uuid="res-002",
            resource_type="bigquery",
            events=events,
        )

        body = client.make_request.call_args.kwargs["body"]
        assert len(body["events"]) == 3

    def test_raises_on_empty_events(self):
        client = _mock_client()
        svc = IngestionService(mc_client=client)

        with self.assertRaises(ValueError):
            svc.send_query_logs(
                resource_uuid="res-003",
                resource_type="snowflake",
                events=[],
            )
        client.make_request.assert_not_called()

    def test_wraps_http_error_as_ingestion_error(self):
        client = _mock_client()
        response = Mock(spec=Response)
        response.text = '{"error": "unauthorized"}'
        client.make_request.side_effect = HTTPError(response=response)

        svc = IngestionService(mc_client=client)

        with self.assertRaises(IngestionError) as ctx:
            svc.send_query_logs(
                resource_uuid="res-004",
                resource_type="snowflake",
                events=[
                    QueryLogEntry(
                        start_time=_dt("2026-03-02T10:00:00Z"),
                        end_time=_dt("2026-03-02T10:00:01Z"),
                        query_text="SELECT 1",
                    ),
                ],
            )
        assert "unauthorized" in str(ctx.exception)


class TestSendLineageRaw(TestCase):
    def test_sends_raw_payload(self):
        client = _mock_client()
        client.make_request.return_value = {"status": "ok"}
        svc = IngestionService(mc_client=client)

        raw = {
            "event_type": "LINEAGE",
            "resource": {"uuid": "r1", "resource_type": "snowflake"},
            "events": [
                {
                    "destination": {
                        "type": "TABLE",
                        "database": "db",
                        "schema": "s",
                        "name": "t",
                    },
                    "sources": [
                        {
                            "type": "TABLE",
                            "database": "db",
                            "schema": "s",
                            "name": "src",
                        },
                    ],
                },
            ],
        }

        result = svc.send_lineage_raw(payload=raw)

        assert result == {"status": "ok"}
        client.make_request.assert_called_once_with(
            path="/ingest/v1/lineage",
            method="POST",
            body=raw,
        )

    def test_wraps_http_error(self):
        client = _mock_client()
        response = Mock(spec=Response)
        response.text = "Bad Request"
        client.make_request.side_effect = HTTPError(response=response)

        svc = IngestionService(mc_client=client)

        with self.assertRaises(IngestionError):
            svc.send_lineage_raw(payload={"event_type": "LINEAGE"})


class TestSendQueryLogsRaw(TestCase):
    def test_sends_raw_payload(self):
        client = _mock_client()
        client.make_request.return_value = {"status": "ok"}
        svc = IngestionService(mc_client=client)

        raw = {
            "event_type": "QUERY_LOG",
            "resource": {"uuid": "r1", "resource_type": "snowflake"},
            "events": [
                {
                    "start_time": "2026-03-02T10:00:00Z",
                    "end_time": "2026-03-02T10:00:05Z",
                    "query_text": "SELECT 1",
                },
            ],
        }

        result = svc.send_query_logs_raw(payload=raw)

        assert result == {"status": "ok"}
        client.make_request.assert_called_once_with(
            path="/ingest/v1/querylogs",
            method="POST",
            body=raw,
        )

    def test_wraps_http_error(self):
        client = _mock_client()
        response = Mock(spec=Response)
        response.text = "Bad Request"
        client.make_request.side_effect = HTTPError(response=response)

        svc = IngestionService(mc_client=client)

        with self.assertRaises(IngestionError):
            svc.send_query_logs_raw(payload={"event_type": "QUERY_LOG"})



#!/usr/bin/env python3
"""
Send query log events to Monte Carlo using the pycarlo IngestionService.

Prerequisites:
    1. Install pycarlo locally:  pip install -e /path/to/python-sdk
    2. Create an integration key with Ingestion scope:
         montecarlo integrations create-key --scope Ingestion
      Or via the GraphQL API / UI.

Usage:
    python sample_ingest_query_logs.py \
        --key-id <INGESTION_KEY_ID> \
        --key-token <INGESTION_KEY_SECRET> \
        --resource-uuid <WAREHOUSE_UUID> \
        --resource-type snowflake \
        --asset-prefix query_logs_run_01
"""

import argparse
import json
import uuid
from datetime import datetime, timedelta, timezone

from pycarlo.core import Client, Session
from pycarlo.features.ingestion import IngestionService
from pycarlo.features.ingestion.models import QueryLogEntry

DEFAULT_ENDPOINT = "https://integrations.getmontecarlo.com"
DEFAULT_ASSET_PREFIX = "run_01"
MANUAL_INGESTION_DATABASE = "manual_ingestion"
MANUAL_INGESTION_SCHEMA = "test"


# Time window for sample events (used for verify search)
SAMPLE_START_TIME = datetime.now(timezone.utc) - timedelta(hours=1)
SAMPLE_END_TIME = datetime.now(timezone.utc) + timedelta(minutes=55)


def asset_name(prefix: str, suffix: str) -> str:
    return f"{prefix}_{suffix}"


def build_table_fqn(prefix: str, suffix: str) -> str:
    return f"{MANUAL_INGESTION_DATABASE}.{MANUAL_INGESTION_SCHEMA}.{asset_name(prefix, suffix)}"


def build_sample_events(prefix: str) -> tuple[list[QueryLogEntry], list[str]]:
    """Build sample events and return (events, query_ids) for verification."""
    now = datetime.now(timezone.utc)
    query_ids = [str(uuid.uuid4()), str(uuid.uuid4()), str(uuid.uuid4())]
    source_a = build_table_fqn(prefix, "a")
    source_b = build_table_fqn(prefix, "b")
    destination_c = build_table_fqn(prefix, "c")
    downstream_view = build_table_fqn(prefix, "view_c")

    print(f"Query IDs: {query_ids}")
    print(
        "Sample assets referenced by query logs: "
        f"{asset_name(prefix, 'a')}, "
        f"{asset_name(prefix, 'b')}, "
        f"{asset_name(prefix, 'c')}, "
        f"{asset_name(prefix, 'view_c')}"
    )

    events = [
        QueryLogEntry(
            start_time=SAMPLE_START_TIME,
            end_time=now - timedelta(minutes=58),
            query_text=(
                f"SELECT id, amount, created_at FROM {source_a} "
                "WHERE created_at >= CURRENT_DATE - INTERVAL '1 day' /**/"
            ),
            query_id=query_ids[0],
            user="analyst@company.com",
            returned_rows=1,
        ),
        QueryLogEntry(
            start_time=now - timedelta(minutes=57),
            end_time=now - timedelta(minutes=56),
            query_text=f"SELECT * FROM {destination_c} WHERE event_type = 'missing' /**/",
            query_id=query_ids[1],
            user="dev@company.com",
            error_code="NO_DATA",
            error_text="No matching rows found for sample query",
        ),
        QueryLogEntry(
            start_time=now - timedelta(minutes=58),
            end_time=now - timedelta(minutes=57),
            query_text=(
                "SELECT c.event_id, c.event_type, b.status "
                f"FROM {destination_c} c "
                f"LEFT JOIN {source_b} b ON c.event_type = b.status "
                "LIMIT 100 /**/"
            ),
            query_id=query_ids[2],
            user="bi_tool@company.com",
            returned_rows=100,
            extra={"warehouse": "COMPUTE_WH", "role": "ANALYST"},
        ),
        QueryLogEntry(
            start_time=now - timedelta(minutes=55),
            end_time=now - timedelta(minutes=54),
            query_text=(
                f"SELECT event_id, event_type FROM {downstream_view} "
                "ORDER BY event_id DESC LIMIT 25 /**/"
            ),
            query_id=str(uuid.uuid4()),
            user="analytics_app@company.com",
            returned_rows=25,
        ),
    ]
    return events, [*query_ids, events[3].query_id]


def main():
    parser = argparse.ArgumentParser(
        description="Send sample query logs to Monte Carlo via pycarlo IngestionService",
    )
    parser.add_argument("--key-id", required=True, help="Ingestion key ID (x-mcd-id)")
    parser.add_argument("--key-token", required=True, help="Ingestion key token (x-mcd-token)")
    parser.add_argument(
        "--endpoint",
        default=DEFAULT_ENDPOINT,
        help=f"Integration Gateway endpoint (default: {DEFAULT_ENDPOINT})",
    )
    parser.add_argument("--resource-uuid", default="your-warehouse-uuid", help="Resource UUID")
    parser.add_argument(
        "--resource-type",
        default="snowflake",
        help="Resource type (e.g. snowflake, bigquery)",
    )
    parser.add_argument(
        "--payload-file",
        help="Path to a JSON file with a custom payload (uses send_query_logs_raw)",
    )
    parser.add_argument(
        "--asset-prefix",
        default=DEFAULT_ASSET_PREFIX,
        help=(
            "Prefix for the sample assets so query logs can target the same fresh "
            f"set used by metadata and lineage (default: {DEFAULT_ASSET_PREFIX})"
        ),
    )
    args = parser.parse_args()

    client = Client(
        session=Session(
            mcd_id=args.key_id,
            mcd_token=args.key_token,
            endpoint=args.endpoint,
            scope="Ingestion",
        )
    )
    service = IngestionService(mc_client=client)

    if args.payload_file:
        with open(args.payload_file) as f:
            payload = json.load(f)
        print(f"Sending raw payload from {args.payload_file} ...")
        result = service.send_query_logs_raw(payload=payload)
        print(f"Response: {json.dumps(result, indent=2) if result else '(empty)'}")
    else:
        events, query_ids = build_sample_events(args.asset_prefix)
        print(f"Sending {len(events)} query log events to resource {args.resource_uuid} ...")
        result = service.send_query_logs(
            resource_uuid=args.resource_uuid,
            resource_type=args.resource_type,
            events=events,
        )
        invocation_id = service.extract_invocation_id(result)
        if invocation_id:
            print(f"Invocation ID: {invocation_id}")
        print(f"Response: {json.dumps(result, indent=2) if result else '(empty)'}")

    print("Done.")


if __name__ == "__main__":
    main()



# Queries related to Monitor execution

RUN_MONITOR = """
mutation runMonitor($monitorUuid: UUID!, $runtimeVariables: [RuntimeVariableValueInput!]) {
    runMonitor(monitorUuid: $monitorUuid, runtimeVariables: $runtimeVariables) {
        success
        manualMonitorExecutionUuid
    }
}
"""

GET_MANUAL_MONITOR_EXECUTION_STATUS = """
query getManualMonitorExecutionStatus($manualMonitorExecutionUuid: UUID!) {
    getManualMonitorExecutionStatus(manualMonitorExecutionUuid: $manualMonitorExecutionUuid) {
        status
        breached
        completed
    }
}
"""

GET_AGENT_MONITOR_MINIMAL = """
query getMonitor($uuids: [String]) {
    getMonitors(uuids: $uuids) {
        uuid
        name
        description
        monitorType
        entityMcons
        agentSpanFilters {
            agent { value }
            spanName { value }
            task { value }
            workflow { value }
        }
        filters {
            type
            operator
            conditions {
                ... on FilterBinary {
                    type
                    predicate { name negated }
                    left { type ... on FilterValueMapKey { key } }
                    right { type ... on FilterValueLiteral { literal } }
                }
            }
        }
    }
}
"""



from __future__ import annotations

from typing import Optional

from requests import HTTPError

from pycarlo.common import get_logger
from pycarlo.common.errors import InvalidSessionError
from pycarlo.core import Client
from pycarlo.features.ingestion.exceptions import IngestionError
from pycarlo.features.ingestion.models import (
    LineageEvent,
    LineageEventType,
    QueryLogEntry,
    RelationalAsset,
    build_lineage_payload,
    build_metadata_payload,
    build_query_log_payload,
)

logger = get_logger(__name__)

_METADATA_PATH = "/ingest/v1/metadata"
_LINEAGE_PATH = "/ingest/v1/lineage"
_QUERY_LOG_PATH = "/ingest/v1/querylogs"


class IngestionService:
    """
    Send observability data to Monte Carlo via the Ingest API.

    The ``Client`` used to initialise this service **must** be created with a
    ``Session`` that has ``scope="Ingestion"`` and credentials for an
    integration key with the *Ingestion* scope.

    Example::

        from pycarlo.core import Client, Session
        from pycarlo.features.ingestion import IngestionService
        from pycarlo.features.ingestion.models import (
            RelationalAsset, AssetMetadata, AssetField,
        )

        client = Client(session=Session(
            mcd_id="<key-id>",
            mcd_token="<key-token>",
            scope="Ingestion",
        ))
        svc = IngestionService(mc_client=client)

        svc.send_metadata(
            resource_uuid="<uuid>",
            resource_type="snowflake",
            events=[
                RelationalAsset(
                    type="TABLE",
                    metadata=AssetMetadata(
                        name="orders",
                        database="analytics",
                        schema="public",
                    ),
                ),
            ],
        )
    """

    def __init__(self, mc_client: Optional[Client] = None):
        """
        :param mc_client: A ``Client`` whose session has ``scope`` set
            (e.g. ``"Ingestion"``).  If omitted a default ``Client()`` is
            created, but it will fail at request time unless the default
            session already carries a scope.
        """
        self._client = mc_client or Client()
        if not self._client.session_scope:
            raise InvalidSessionError(
                "IngestionService requires a Client with scope set "
                '(e.g. Session(scope="Ingestion")).'
            )

    def send_metadata(
        self,
        resource_uuid: str,
        resource_type: str,
        events: list[RelationalAsset],
    ) -> dict | None:
        """
        Send relational-asset metadata to Monte Carlo.

        :param resource_uuid: UUID of the Monte Carlo resource (warehouse/lake).
        :param resource_type: Resource type identifier, e.g. ``"snowflake"``,
            ``"bigquery"`` (lowercase).
        :param events: One or more :class:`RelationalAsset` objects describing
            the tables/views to ingest.
        :return: The JSON response from the API, or ``None`` if the response
            body was empty.
        :raises IngestionError: If the API returns an HTTP error.
        """
        if not events:
            raise ValueError("At least one RelationalAsset event is required.")

        payload = build_metadata_payload(
            resource_uuid=resource_uuid,
            resource_type=resource_type,
            events=events,
        )
        return self._post_metadata(payload)

    @staticmethod
    def extract_invocation_id(response: dict | None) -> str | None:
        """
        Extract the invocation ID returned by the ingest API.

        The Integration Gateway returns ``{"invocation_id": "<uuid>"}`` for
        successful ingest metadata, query-log, and lineage requests. This helper keeps SDK
        callers from needing to reach into the raw response payload directly.

        :param response: The JSON response returned by one of the public send
            methods on this service.
        :return: The invocation ID when present, otherwise ``None``.
        """
        if not isinstance(response, dict):
            return None

        invocation_id = response.get("invocation_id")
        return invocation_id if isinstance(invocation_id, str) else None

    def send_metadata_raw(self, payload: dict) -> dict | None:
        """
        Send a raw metadata payload dictionary to the ingest API.

        Use this when you already have a pre-built payload that conforms to the
        ``POST /ingest/v1/metadata`` schema.

        :param payload: The full request body as a dictionary.
        :return: The JSON response from the API, or ``None``.
        :raises IngestionError: If the API returns an HTTP error.
        """
        return self._post_metadata(payload)

    # ------------------------------------------------------------------
    # Lineage
    # ------------------------------------------------------------------

    def send_lineage(
        self,
        resource_uuid: str,
        resource_type: str,
        events: list[LineageEvent],
        event_type: LineageEventType | str | None = None,
    ) -> dict | None:
        """
        Send lineage data to Monte Carlo.

        :param resource_uuid: UUID of the Monte Carlo resource (warehouse/lake).
        :param resource_type: Resource type identifier, e.g. ``"snowflake"``,
            ``"bigquery"`` (lowercase).
        :param events: One or more :class:`LineageEvent` objects describing
            the data-flow relationships to ingest.
        :param event_type: Explicit event type (``"LINEAGE"`` or
            ``"COLUMN_LINEAGE"``).  When *None* the type is auto-detected:
            ``"COLUMN_LINEAGE"`` if any event contains ``fields``, otherwise
            ``"LINEAGE"``.
        :return: The JSON response from the API, or ``None`` if the response
            body was empty.
        :raises IngestionError: If the API returns an HTTP error.
        """
        if not events:
            raise ValueError("At least one LineageEvent is required.")

        payload = build_lineage_payload(
            resource_uuid=resource_uuid,
            resource_type=resource_type,
            events=events,
            event_type=event_type,
        )
        return self._post_lineage(payload)

    def send_lineage_raw(self, payload: dict) -> dict | None:
        """
        Send a raw lineage payload dictionary to the ingest API.

        Use this when you already have a pre-built payload that conforms to the
        ``POST /ingest/v1/lineage`` schema.

        :param payload: The full request body as a dictionary.
        :return: The JSON response from the API, or ``None``.
        :raises IngestionError: If the API returns an HTTP error.
        """
        return self._post_lineage(payload)

    # ------------------------------------------------------------------
    # Query logs
    # ------------------------------------------------------------------

    def send_query_logs(
        self,
        resource_uuid: str,
        resource_type: str,
        events: list[QueryLogEntry],
    ) -> dict | None:
        """
        Send query log events to Monte Carlo.

        :param resource_uuid: UUID of the Monte Carlo resource (warehouse/lake).
        :param resource_type: Resource type identifier, e.g. ``"snowflake"``,
            ``"bigquery"`` (lowercase).
        :param events: One or more :class:`QueryLogEntry` objects describing
            the queries to ingest.
        :return: The JSON response from the API, or ``None`` if the response
            body was empty.
        :raises IngestionError: If the API returns an HTTP error.
        """
        if not events:
            raise ValueError("At least one QueryLogEntry event is required.")

        payload = build_query_log_payload(
            resource_uuid=resource_uuid,
            resource_type=resource_type,
            events=events,
        )
        return self._post_query_logs(payload)

    def send_query_logs_raw(self, payload: dict) -> dict | None:
        """
        Send a raw query log payload dictionary to the ingest API.

        Use this when you already have a pre-built payload that conforms to the
        ``POST /ingest/v1/querylogs`` schema.

        :param payload: The full request body as a dictionary.
        :return: The JSON response from the API, or ``None``.
        :raises IngestionError: If the API returns an HTTP error.
        """
        return self._post_query_logs(payload)

    # ------------------------------------------------------------------
    # Internal helpers
    # ------------------------------------------------------------------

    def _post_metadata(self, payload: dict) -> dict | None:
        return self._post(
            path=_METADATA_PATH,
            payload=payload,
            label="Metadata",
        )

    def _post_lineage(self, payload: dict) -> dict | None:
        return self._post(
            path=_LINEAGE_PATH,
            payload=payload,
            label="Lineage",
        )

    def _post(self, path: str, payload: dict, label: str) -> dict | None:
        try:
            return self._client.make_request(
                path=path,
                method="POST",
                body=payload,
            )
        except HTTPError as exc:
            response_body = ""
            if exc.response is not None:
                try:
                    response_body = exc.response.text
                except Exception:
                    pass
            raise IngestionError(
                f"{label} ingestion request failed: {exc}. Response: {response_body}"
            ) from exc

    def _post_query_logs(self, payload: dict) -> dict | None:
        return self._post(
            path=_QUERY_LOG_PATH,
            payload=payload,
            label="Query log",
        )
